@@ -7,7 +7,9 @@
 # http://opensource.org/licenses/MIT>, at your option. This file may not be
 # copied, modified, or distributed except according to those terms.
 
+import base64
 import getpass
+import hashlib
 import os
 import sys
 import time
@@ -16,6 +18,7 @@ import click
 from cryptography.hazmat.primitives import hashes
 from fido2.client import ClientError as Fido2ClientError
 from fido2.ctap1 import ApduError
+from fido2.ctap2 import CredentialManagement
 
 import solo
 import solo.fido2
@@ -29,6 +32,12 @@ def key():
     pass
 
 
+@click.group(name="credential")
+def cred():
+    """Credential management, see subcommands."""
+    pass
+
+
 @click.group()
 def rng():
     """Access TRNG on key, see subcommands."""
@@ -73,8 +82,8 @@ def feedkernel(count, serial):
 
     p = solo.client.find(serial)
 
-    import struct
     import fcntl
+    import struct
 
     RNDADDENTROPY = 0x40085203
 
@@ -115,6 +124,7 @@ def feedkernel(count, serial):
     "--host", help="Relying party's host", default="solokeys.dev", show_default=True
 )
 @click.option("--user", help="User ID", default="they", show_default=True)
+@click.option("--pin", help="PIN", default=None)
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
@@ -124,7 +134,7 @@ def feedkernel(count, serial):
     default="Touch your authenticator to generate a credential...",
     show_default=True,
 )
-def make_credential(serial, host, user, udp, prompt):
+def make_credential(serial, host, user, udp, prompt, pin):
     """Generate a credential.
 
     Pass `--prompt ""` to output only the `credential_id` as hex.
@@ -132,8 +142,20 @@ def make_credential(serial, host, user, udp, prompt):
 
     import solo.hmac_secret
 
+    # check for PIN
+    if not pin:
+        pin = getpass.getpass("PIN (leave empty for no PIN): ")
+    if not pin:
+        pin = None
+
     solo.hmac_secret.make_credential(
-        host=host, user_id=user, serial=serial, output=True, prompt=prompt, udp=udp
+        host=host,
+        user_id=user,
+        serial=serial,
+        output=True,
+        prompt=prompt,
+        udp=udp,
+        pin=pin,
     )
 
 
@@ -141,6 +163,7 @@ def make_credential(serial, host, user, udp, prompt):
 @click.option("-s", "--serial", help="Serial number of Solo use")
 @click.option("--host", help="Relying party's host", default="solokeys.dev")
 @click.option("--user", help="User ID", default="they")
+@click.option("--pin", help="PIN", default=None)
 @click.option(
     "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
 )
@@ -152,7 +175,7 @@ def make_credential(serial, host, user, udp, prompt):
 )
 @click.argument("credential-id")
 @click.argument("challenge")
-def challenge_response(serial, host, user, prompt, credential_id, challenge, udp):
+def challenge_response(serial, host, user, prompt, credential_id, challenge, udp, pin):
     """Uses `hmac-secret` to implement a challenge-response mechanism.
 
     We abuse hmac-secret, which gives us `HMAC(K, hash(challenge))`, where `K`
@@ -169,6 +192,12 @@ def challenge_response(serial, host, user, prompt, credential_id, challenge, udp
 
     import solo.hmac_secret
 
+    # check for PIN
+    if not pin:
+        pin = getpass.getpass("PIN (leave empty for no PIN): ")
+    if not pin:
+        pin = None
+
     solo.hmac_secret.simple_secret(
         credential_id,
         challenge,
@@ -178,6 +207,7 @@ def challenge_response(serial, host, user, prompt, credential_id, challenge, udp
         prompt=prompt,
         output=True,
         udp=udp,
+        pin=pin,
     )
 
 
@@ -317,20 +347,25 @@ def set_pin(serial):
 def verify(pin, serial, udp):
     """Verify key is valid Solo Secure or Solo Hacker."""
 
+    key = solo.client.find(serial, udp=udp)
+
+    if (
+        key.client
+        and ("clientPin" in key.client.info.options)
+        and key.client.info.options["clientPin"]
+        and not pin
+    ):
+        pin = getpass.getpass("PIN: ")
+
     # Any longer and this needs to go in a submodule
     print("Please press the button on your Solo key")
     try:
-        cert = solo.client.find(serial, udp=udp).make_credential(pin=pin)
-    except ValueError as e:
-        # python-fido2 library pre-emptively returns `ValueError('PIN required!')`
-        # instead of trying, and returning  `CTAP error: 0x36 - PIN_REQUIRED`
-        if "PIN required" in str(e):
-            print("Your key has a PIN set. Please pass it using `--pin <your PIN>`")
-            sys.exit(1)
-        raise
-
+        cert = key.make_credential(pin=pin)
     except Fido2ClientError as e:
         cause = str(e.cause)
+        if "PIN required" in cause:
+            print("Your key has a PIN set but none was provided.")
+            sys.exit(1)
         # error 0x31
         if "PIN_INVALID" in cause:
             print("Your key has a different PIN. Please try to remember it :)")
@@ -453,6 +488,20 @@ def ping(serial, udp, ping_data):
     print(f"took {duration} ms")
 
 
+@click.command()
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+@click.argument("sequence")
+def keyboard(serial, sequence):
+    """Program the specified key sequence to Solo"""
+
+    dev = solo.client.find(serial)
+    buf = sequence.encode("ascii")
+    if len(buf) > 64:
+        print("Keyboard sequence cannot exceed 64 bytes")
+    else:
+        dev.program_kbd(buf)
+
+
 @click.command()
 @click.option("-s", "--serial", help="Serial number of Solo to use")
 def disable_updates(serial):
@@ -469,6 +518,105 @@ def disable_updates(serial):
         print("Failed to disable the firmware update.")
 
 
+@click.command(name="info")
+@click.option("--pin", help="PIN for to access key")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+@click.option(
+    "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
+)
+def cred_info(pin, serial, udp):
+    """Get credentials metadata"""
+    if not pin:
+        pin = getpass.getpass("PIN: ")
+
+    client = solo.client.find(serial, udp=udp)
+    cm = client.cred_mgmt(pin)
+    meta = cm.get_metadata()
+    existing = meta[CredentialManagement.RESULT.EXISTING_CRED_COUNT]
+    remaining = meta[CredentialManagement.RESULT.MAX_REMAINING_COUNT]
+    print("Existing resident keys: {}".format(existing))
+    print("Remaining resident keys: {}".format(remaining))
+
+
+@click.command(name="ls")
+@click.option("--pin", help="PIN for to access key")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+@click.option(
+    "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
+)
+def cred_ls(pin, serial, udp):
+    """List stored credentials"""
+    if not pin:
+        pin = getpass.getpass("PIN: ")
+
+    client = solo.client.find(serial, udp=udp)
+    cm = client.cred_mgmt(pin)
+    meta = cm.get_metadata()
+    existing = meta[CredentialManagement.RESULT.EXISTING_CRED_COUNT]
+    if existing == 0:
+        print("No resident credentials on this device.")
+        return
+    rps = cm.enumerate_rps()
+    all_creds = {}
+    for rp in rps:
+        rp_id = rp[CredentialManagement.RESULT.RP]["id"]
+        creds = cm.enumerate_creds(rp[CredentialManagement.RESULT.RP_ID_HASH])
+        all_creds[rp_id] = creds
+    if all_creds:
+        print("{:20}{:20}{}".format("Relying Party", "Username", "Credential ID"))
+        print("-" * 53)
+    for rp_id, creds in all_creds.items():
+        for cred in creds:
+            user = cred.get(CredentialManagement.RESULT.USER, "")
+            cred_id = cred[CredentialManagement.RESULT.CREDENTIAL_ID]["id"]
+            cred_id_b64 = base64.b64encode(cred_id).decode("ascii")
+            print("{:20}{:20}{}".format(rp_id, user["name"], cred_id_b64))
+
+
+@click.command(name="rm")
+@click.option("--pin", help="PIN for to access key")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+@click.option(
+    "--udp", is_flag=True, default=False, help="Communicate over UDP with software key"
+)
+@click.argument("credential-id")
+def cred_rm(pin, credential_id, serial, udp):
+    """Remove stored credential"""
+    if not pin:
+        pin = getpass.getpass("PIN: ")
+
+    client = solo.client.find(serial, udp=udp)
+    cm = client.cred_mgmt(pin)
+    cred = {"id": base64.b64decode(credential_id), "type": "public-key"}
+    cm.delete_cred(cred)
+
+
+@click.command()
+@click.option("--pin", help="PIN for to access key")
+@click.option("-s", "--serial", help="Serial number of Solo to use")
+@click.argument("credential-id")
+@click.argument("filename")
+def sign_file(pin, serial, credential_id, filename):
+    """Sign the specified file using the given credential-id"""
+
+    dev = solo.client.find(serial)
+    dgst = hashlib.sha256()
+    with open(filename, "rb") as f:
+        while True:
+            data = f.read(64 * 1024)
+            if not data:
+                break
+            dgst.update(data)
+    print("{0}  {1}".format(dgst.hexdigest(), filename))
+    print("Please press the button on your Solo key")
+    ret = dev.sign_hash(base64.b64decode(credential_id), dgst.digest(), pin)
+    sig = ret[1]
+    sig_file = filename + ".sig"
+    print("Saving signature to " + sig_file)
+    with open(sig_file, "wb") as f:
+        f.write(sig)
+
+
 key.add_command(rng)
 rng.add_command(hexbytes)
 rng.add_command(raw)
@@ -487,3 +635,9 @@ key.add_command(verify)
 key.add_command(wink)
 key.add_command(disable_updates)
 key.add_command(ping)
+key.add_command(keyboard)
+key.add_command(cred)
+key.add_command(sign_file)
+cred.add_command(cred_info)
+cred.add_command(cred_ls)
+cred.add_command(cred_rm)
